#include <iostream>
#include <stdexcept>

#include <boost/foreach.hpp>

#include "configuration.hpp"

namespace application {
namespace {

  std::string resolveString(boost::property_tree::ptree& a_ptree,
                            const char* a_field_name,
                            bool a_valid_empty,
                            bool a_empty_as_default,
                            bool a_valid_null,
                            std::string a_default
                            ){
    bool exists = a_ptree.find(a_field_name) != a_ptree.not_found();
    if (!exists && a_valid_null)
      return a_default;
    else if (!exists)
      throw std::runtime_error(std::string(a_field_name) + " config parameter not set");
    std::string result = a_ptree.get<std::string>(a_field_name);
    if (result.empty() && a_valid_empty && a_empty_as_default)
      return a_default;
    else if (result.empty() && a_valid_empty)
      return result;
    else if (result.empty() && !a_valid_empty)
      throw std::runtime_error(std::string(a_field_name) + " config parameter is empty");
    return result;
  }

  int resolveInteger( boost::property_tree::ptree& a_ptree,
                      const char* a_field_name,
                      bool a_valid_null,
                      int  a_min,
                      int  a_max,
                      int a_default
                            ){
    bool exists = a_ptree.find(a_field_name) != a_ptree.not_found();
    if (!exists && a_valid_null)
      return a_default;
    else if (!exists)
      throw std::runtime_error(std::string(a_field_name) + " config parameter not set");
    int result = 0;
    try {
      result = a_ptree.get<int>(a_field_name);
    } catch(const std::exception&){
      throw std::runtime_error(std::string("The ") + std::string(a_field_name) + " parameter must be of the integer type");
    }
    if (a_min != -1 && result < a_min){
      throw std::runtime_error(std::string("The ") + std::string(a_field_name) + " configuration parameter must be greater than " + std::to_string(a_min));
    }
    if (a_max != -1 && result > a_max){
      throw std::runtime_error(std::string("The ") + std::string(a_field_name) + " configuration parameter must be less than " + std::to_string(a_max));
    }
    return result;
  }

  int resolveBoolean( boost::property_tree::ptree& a_ptree,
                      const char* a_field_name,
                      int a_default
                      ){
    bool exists = a_ptree.find(a_field_name) != a_ptree.not_found();
    if (!exists)
      return a_default;
    bool result = false;
    try {
      result = a_ptree.get<bool>(a_field_name);
    } catch(const std::exception&){
      throw std::runtime_error(std::string("The ") + std::string(a_field_name) + " parameter must be of the boolean type");
    }
    return result;
  }  

  void fillPorts(std::vector<int>& a_vector, int a_count){
    int lastPort = a_vector[a_vector.size()-1];
    for(int i = (int)a_vector.size(); i < a_count; ++i){
      ++lastPort;
      a_vector.push_back(lastPort);
    }
  }


} // non namespace


  void handler_configuration_t::set_values(boost::property_tree::ptree& a_ptree, endpoint_configuration_t& a_owner) {
    this->serverControlPort       = a_owner.serverControlPort;
    this->user                    = resolveString(a_ptree, "user", true, true, true, a_owner.user);
    this->password                = resolveString(a_ptree, "password", true, true, true, a_owner.password);
    this->listenQueue             = resolveInteger(a_ptree, "listenQueue", true, 1, -1, a_owner.listenQueue);
    this->clientSocketTimeout     = resolveInteger(a_ptree, "clientSocketTimeout", true, 0, -1, a_owner.clientSocketTimeout);
    this->maxRequestTimeout       = resolveInteger(a_ptree, "maxRequestTimeout", true, 0, -1, a_owner.maxRequestTimeout);
    this->serverSocketTimeout     = resolveInteger(a_ptree, "serverSocketTimeout", true, 0, -1, a_owner.serverSocketTimeout);
    this->maxResponseTimeout      = resolveInteger(a_ptree, "maxResponseTimeout", true, 0, -1, a_owner.maxResponseTimeout);
    this->startTimeout            = resolveInteger(a_ptree, "startTimeout", true, 0, -1, a_owner.startTimeout);
    this->stopTimeout             = resolveInteger(a_ptree, "stopTimeout", true, 0, -1, a_owner.stopTimeout);
    this->keepAliveTimeout        = resolveInteger(a_ptree, "keepAliveTimeout", true, 0, -1, a_owner.keepAliveTimeout);

    this->host                    = resolveString(a_ptree, "host", true, true, true, this->host);
    this->path                    = resolveString(a_ptree, "path", true, true, true, this->path);
    this->processes               = resolveInteger(a_ptree, "processes", false, 1, -1, 0);
    this->processesLeakProtection = resolveInteger(a_ptree, "processesLeakProtection", false, 1, -1, 0);
    if (!this->processes && !this->processesLeakProtection)
      throw std::runtime_error("The total number of processes of the handler is zero");

    this->script                  = resolveString(a_ptree, "script", false, false, false, "");
    this->maxMemory               = resolveInteger(a_ptree, "maxMemory", true, 0, -1, 0);
    this->maxMemoryLeakProtection = resolveInteger(a_ptree, "maxMemoryLeakProtection", true, 0, -1, 0);

    if (a_ptree.find("dataPorts") == a_ptree.not_found())
      throw std::runtime_error("dataPorts config parameter not set");
    boost::property_tree::ptree ptDataPorts = a_ptree.get_child("dataPorts");
    this->dataPorts.clear();
    try {
      BOOST_FOREACH(auto &p, ptDataPorts){
        this->dataPorts.push_back(std::stoi(p.second.data()));
      }
    } catch(const std::exception&) {
      throw std::runtime_error("The dataPorts config parameter must be an array of integers");
    }
    if (this->dataPorts.empty()) {
      throw std::runtime_error("The dataPorts parameter cannot be empty");
    }
    fillPorts(this->dataPorts, (this->processes + this->processesLeakProtection) * 2);

    if (a_ptree.find("controlPorts") == a_ptree.not_found())
      throw std::runtime_error("controlPorts config parameter not set");
    boost::property_tree::ptree ptControlPorts = a_ptree.get_child("controlPorts");
    this->controlPorts.clear();
    try {
      BOOST_FOREACH(auto &p, ptControlPorts) {
        this->controlPorts.push_back(std::stoi(p.second.data()));
      }
    } catch(const std::exception&) {
      throw std::runtime_error("The controlPorts config parameter must be an array of integers");
    }
    if (this->controlPorts.empty()) {
      throw std::runtime_error("The controlPorts parameter cannot be empty");
    }
    fillPorts(this->controlPorts, (this->processes + this->processesLeakProtection) * 2);

    this->nodeOptions.clear();
    if (a_ptree.find("nodeOptions") != a_ptree.not_found()) {
      boost::property_tree::ptree ptNodeOptions = a_ptree.get_child("nodeOptions");
      try {
        BOOST_FOREACH(auto &ptNodeOptionsItem, ptNodeOptions) {
          this->nodeOptions.push_back(std::vector<std::string>());
          BOOST_FOREACH(auto &p, ptNodeOptionsItem.second) {
            this->nodeOptions[this->nodeOptions.size()-1].push_back(p.second.data());
          }
        }
      } catch(const std::exception&) {
        throw std::runtime_error("The nodeOptions config parameter must be an array of strings");
      }
    }
    for(size_t i = this->nodeOptions.size(); i < this->processes*2 + this->processesLeakProtection*2; ++i)
      this->nodeOptions.push_back(std::vector<std::string>());
  }

  void endpoint_configuration_t::set_values(boost::property_tree::ptree& a_ptree, server_configuration_t& a_owner){
    this->serverControlPort   = a_owner.serverControlPort;
    this->user                = resolveString(a_ptree, "user", true, true, true, a_owner.user);
    this->password            = resolveString(a_ptree, "password", true, true, true, a_owner.password);
    this->listenQueue         = resolveInteger(a_ptree, "listenQueue", true, 1, -1, a_owner.listenQueue);
    this->clientSocketTimeout = resolveInteger(a_ptree, "clientSocketTimeout", true, 0, -1, a_owner.clientSocketTimeout);
    this->maxRequestTimeout   = resolveInteger(a_ptree, "maxRequestTimeout", true, 0, -1, a_owner.maxRequestTimeout);
    this->serverSocketTimeout = resolveInteger(a_ptree, "serverSocketTimeout", true, 0, -1, a_owner.serverSocketTimeout);
    this->maxResponseTimeout  = resolveInteger(a_ptree, "maxResponseTimeout", true, 0, -1, a_owner.maxResponseTimeout);
    this->startTimeout        = resolveInteger(a_ptree, "startTimeout", true, 0, -1, a_owner.startTimeout);
    this->stopTimeout         = resolveInteger(a_ptree, "stopTimeout", true, 0, -1, a_owner.stopTimeout);
    this->keepAliveTimeout    = resolveInteger(a_ptree, "keepAliveTimeout", true, 0, -1, a_owner.keepAliveTimeout);
    this->bind4               = resolveString(a_ptree, "bind4", true, false, true, this->bind4);
    this->bind6               = resolveString(a_ptree, "bind6", true, true, true, this->bind6);
    this->port                = resolveInteger(a_ptree, "port", false, 1, 65535, this->port);

    if (a_ptree.find("handlers") == a_ptree.not_found())
      throw std::runtime_error("handlers config parameter not set");
    boost::property_tree::ptree handlers = a_ptree.get_child("handlers");
    if (handlers.size() == 0)
      throw std::runtime_error("handlers config parameter must contain an array of objects");
    BOOST_FOREACH(auto &h, handlers){
      handler_configuration_t hc;
      hc.set_values(h.second, *this);
      this->handlers.push_back(hc);
    }
  }


  void server_configuration_t::set_values(boost::property_tree::ptree& a_ptree){
    this->logFile             = resolveString(a_ptree, "logFile", true, true, true, std::string(""));    
    std::string logLevel      = resolveString(a_ptree, "logLevel", true, true, true, std::string("log"));
    if      (logLevel == "err") this->logLevel = fcf::LOG_LEVEL_ERR;
    else if (logLevel == "wrn") this->logLevel = fcf::LOG_LEVEL_WRN;
    else if (logLevel == "trc") this->logLevel = fcf::LOG_LEVEL_TRC;
    else                        this->logLevel = fcf::LOG_LEVEL_LOG;
    this->logLifeTime         = resolveInteger(a_ptree, "logLifeTime", true, 0, -1, this->logLifeTime);
    this->serverControlPort   = resolveInteger(a_ptree, "serverControlPort", true, 1, 65535, this->serverControlPort);

    this->user                = resolveString(a_ptree, "user", true, true, true, this->user);
    this->password            = resolveString(a_ptree, "password", true, true, true, this->password);
    this->slave               = resolveBoolean(a_ptree, "slave", this->slave);
    this->listenQueue         = resolveInteger(a_ptree, "listenQueue", true, 1, -1, this->listenQueue);
    this->clientSocketTimeout = resolveInteger(a_ptree, "clientSocketTimeout", true, 0, -1, this->clientSocketTimeout);
    this->maxRequestTimeout   = resolveInteger(a_ptree, "maxRequestTimeout", true, 0, -1, this->maxRequestTimeout);
    this->serverSocketTimeout = resolveInteger(a_ptree, "serverSocketTimeout", true, 0, -1, this->serverSocketTimeout);
    this->maxResponseTimeout  = resolveInteger(a_ptree, "maxResponseTimeout", true, 0, -1, this->maxResponseTimeout);
    this->startTimeout        = resolveInteger(a_ptree, "startTimeout", true, 0, -1, this->startTimeout);
    this->stopTimeout         = resolveInteger(a_ptree, "stopTimeout", true, 0, -1, this->stopTimeout);
    this->keepAliveTimeout    = resolveInteger(a_ptree, "keepAliveTimeout", true, 0, -1, this->keepAliveTimeout);

    if (a_ptree.find("endpoints") == a_ptree.not_found())
      throw std::runtime_error("endpoints config parameter not set");
    boost::property_tree::ptree ptendpoints = a_ptree.get_child("endpoints");
    if (ptendpoints.size() == 0)
      throw std::runtime_error("endpoints config parameter must contain an array of objects");
    BOOST_FOREACH(auto &ep, ptendpoints){
      endpoint_configuration_t epc;
      epc.set_values(ep.second, *this);
      endpoints.push_back(epc);
    }
  }


}

